#!/usr/bin/perl -w

package RPM::Source::Convert::Generic2ALT;
use File::Basename;
use File::Path qw(make_path remove_tree);
use File::Temp qw/tempfile tempdir/;
use File::Spec;
use Cwd;
use strict;
use warnings;
use Carp;

#use RPM::Source::Convert;
use RPM::Source::Convert::DistroMap::Groups;
use RPM::Source::Convert::DistroMap::Packages;
use RPM::Source::Convert::DistroMap::SubpackageRename;

$DistroMap::default_originrepo='fedora';
$DistroMap::default_originbranch=undef;
$DistroMap::default_destrepo='altlinux';
$DistroMap::default_destbranch=undef;

push @RPM::Source::Convert::PREHOOK_FIXES, \&alt_generic_fix;
unshift @RPM::Source::Convert::POSTHOOK_FIXES, \&alt_generic_posthook_fix;


# todo: all is fedora-specific; move
map {$RPM::Source::Convert::DistroMap::Packages::DEP_DROP_VER{$_} = 1} qw!
libjpeg-devel
!;

map {$RPM::Source::Convert::DistroMap::Packages::DEP_SKIP{$_} = 1} qw!
redhat-lsb
/sbin/install-info
arts-devel
emacs(bin)
fontpackages-filesystem
!;
# emacs(bin) -- fc specific...

my %DEP_SUBST = (
'%{_sbindir}/update-alternatives' => 'alternatives',
'/usr/sbin/update-alternatives' => 'alternatives',
'%{_sbindir}/alternatives' => 'alternatives',
'/usr/sbin/alternatives' => 'alternatives',
#'php-devel' => 'php5-devel',
#'php-embedded-devel' => 'php5-devel',
'tclx-devel' => 'tclx',
' ' => ' ',
);
foreach my $key (keys(%DEP_SUBST)) {
    $RPM::Source::Convert::DistroMap::Packages::DEP_SUBST{$key} = $DEP_SUBST{$key};
}

sub alt_generic_posthook_fix {
    my ($self,$spec,$parent)=@_;
    my @sections=$spec->get_sections();

    foreach my $sec (@sections) {
	my $t = $sec->get_type();
	if ($t eq 'package') {
	    &alt_generic_merge_main_group_if_group_is_missing($spec,$sec);
	}
    }

    # merge language-translated summary and description
    if ($parent) {
	my $need_reset=0;
	@sections=$parent->get_sections();
	foreach my $sec (@sections) {
	    my $t = $sec->get_type();
	    my $p = $sec->get_raw_package();
	    my $extra=$sec->get_extra_qualifier();
	    if ($t eq 'package') {
		foreach (@{$sec->get_bodyref()}) {
		    if (/^(Summary\(\S+\)):\s*(.+)/i) {
			my $tag=$1;
			my $val=$2;
			my $newsec=$spec->get_section($t,$p);
			if ($newsec) {
			    my $summary=$newsec->get_tag($tag);
			    $newsec->set_tag($tag,$val) unless $summary;
			}
		    }
		}
	    } elsif ($t eq 'description' and $extra) {
		my $newsec=$spec->get_section($t,$p,$extra);
		unless ($newsec) {
		    $newsec=$spec->get_section($t,$p);
		    next unless $newsec;
		    push @{$newsec->get_bodyref()}, "\n", @{$sec->get_bodyref()};
		    $need_reset=1;
		}
	    }
	}
	$spec->_reset_speclist() if $need_reset;
    }

}


sub alt_generic_fix {
    my ($self,$spec,$parent)=@_;
    my @sections=$spec->get_sections();
    my $cleanroot_pattern='^\s*(rm|%{__rm}) -rf (\$RPM_BUILD_ROOT|\%\{?rpm_build_root\}?|\%\{?buildroot\}?)\s*$';
    &alt_generic_check_ahead($spec,$sections[0]);
    &alt_generic_check_cxx($self,$spec);
    &alt_generic_check_zip($spec);
    &alt_generic_check_jni($spec);
    my $mainsec;
    my %ADD_TO_BR;
    foreach my $sec (@sections) {
	my $t = $sec->get_type();
	if ($t eq 'package') {
	    $sec->exclude('^(Packager|Vendor|Distribution|Build[Rr]oot):');
	    # encoding mismatch. alt assume 8 bit
	    # TODO: detect utf8?
	    #$sec->exclude('^Summary\(de\):');
	    &alt_generic_cleanup_headers($self,$spec,$sec);
	} elsif ($t eq 'description') {
	    my $l=$sec->get_flag('-l');
	    #$sec->delete() if $l and $l=~/^(?:de|fr)$/;
	    &alt_generic_fix_desc($spec, $sec) unless $l;
	} elsif ($t eq 'files') {
	    &alt_generic_cleanup_files($sec);
	} elsif ($t eq 'prep' || $t eq 'build' || $t eq 'install') {
	    $sec->exclude($cleanroot_pattern);
	    if ($t eq 'build') {
		$ADD_TO_BR{'python'}='# fontforge scripts hack: python core modules are required (#25979)'
		    if $sec->match_body('fontforge\s.+-script\s');
	    }
	} elsif ($t eq 'clean') {
	    $sec->exclude($cleanroot_pattern);
	    $sec->delete();
	} elsif ($t eq 'check') {
	    # %check now supported in alt rpm 5.1+
	} elsif ($t eq 'changelog') {
	    unless ($spec->{'CONVERT::GENERIC::NOINFO'}) {
		my ($chfh, $chfilename) = tempfile();
		print $chfh 'EVR: ',scalar $spec->get_evr(),"\n";
		my $mainsec=$spec->get_section('package','');
		for my $tag (qw/Name Version Release Serial Epoch/) {
		    my $value = $mainsec->get_tag($tag);
		    print $chfh $tag, ': ',$value,"\n" if defined $value;
		}
		if ($spec->{'SOURCERPM'}) {
		    print $chfh 'srpm: ',$spec->{'SOURCERPM'},"\n";
		} else {
		    print $chfh 'spec: ',$spec->{'SPECFILE'},"\n";
		}
		print $chfh @{$sec->get_bodyref()};
		close ($chfh);
		$spec->add_source($chfilename, NAME=>'import.info', FILE=>$chfilename);
		unlink $chfilename;
	    }
	    $sec->empty_body();
	} elsif ($t eq 'pre') {
	    &alt_generic_fix_scripts($spec,$sec);
	} elsif ($t eq 'post') {
	    &alt_generic_convert_alternatives($spec,$sec);
	    &alt_generic_fix_scripts($spec,$sec);
	} elsif (($t eq 'postun') or ($t eq 'preun')) {
	    &alt_generic_convert_alternatives($spec,$sec);
	    &alt_generic_fix_scripts($spec,$sec);
	} elsif (($t eq 'triggerin') or ($t eq 'triggerun') or ($t eq 'triggerpostin') or ($t eq 'triggerposun')) {
	    # not good, no state :(
	    #&alt_generic_convert_alternatives($spec,$sec);
	} elsif ($t eq 'pretrans') {
	    # what to do ???
	} elsif ($t eq 'posttrans') {
	    &alt_generic_fix_scripts($spec,$sec);
	}
    }
    if (%ADD_TO_BR) {
	$sections[0]->unshift_body(
	    join("\n",grep {/^#/} values(%ADD_TO_BR))."\n".
	    "BuildRequires: ".join(' ',keys %ADD_TO_BR)."\n"
	    );
    }
}

sub alt_generic_cleanup_headers {
    my ($self,$spec,$section)=@_;
    my @out;
    my $is_prev_line_multiline=0;
    foreach (@{$section->get_bodyref()}) {
	if (/^(url|source\d*):(\s*)(.+\S)\s*$/i) {
	    my $Tag=$1;
	    my $tspace=$2;
	    my $url = $3;
	    # TODO: beware of macroses, like Source2: blabla-%{ename}-%bversion.zip

	    # unquoted space in url hack (common in fedora)
	    $url=~s/(?<!\%)\%(2[0-9])/\%\%$1/g;
	    # unquoted unicode in project url hack
	    $url=~s/(?<!\%)\%(?=[ABCDEF][0-9])/\%\%/g if $Tag=~/^url$/i;
	    push @out, "$Tag:$tspace$url\n";
	} elsif (/^release:.*\%mkrel/i) {
	    s/\%mkrel\s+//;
	    $spec->macros()->{release}=~s/\%mkrel\s+// if $spec->macros()->{release};
	    push @out, $_;
	} elsif (/^summary:/i) {
	    push @out, &__kill_non_ascii($spec, $_);
	} elsif (/^\s*#/i) {
	    #in comment: conakry-fonts-20070829-4.fc15/*.spec does not look like a spec file
	    push @out, &__kill_non_ascii($spec, $_);
	} elsif (/\\\n$/ or $is_prev_line_multiline) {
	    # multiline macro for description: fonts-otf-beteckna
	    push @out, &__kill_non_ascii($spec, $_);
	} else {
	    push @out, $_;
	}
	$is_prev_line_multiline= /\\\n$/;
    }
    $section->set_body(\@out);
}

# is macro %name used in spec before Name:,Version: tags
# it is not allowed in alt
sub alt_generic_check_ahead {
    my ($spec,$section)=@_;
    my @out;
    my ($is_name_passed,$is_name_ahead);
    my ($is_version_passed,$is_version_ahead);
    foreach (@{$section->get_bodyref()}) {
	m/^\s*name:/i and $is_name_passed=1; #and print STDERR "passed $_\n";
	m/^\s*version:/i and $is_version_passed=1; #and print STDERR "passed $_\n";
	$is_name_ahead=1 if not $is_name_passed and (m/\%\{name\}/ or m/\%name(?:[^\w\d_]|$)/);
	$is_version_ahead=1 if not $is_version_passed and (m/\%\{version\}/ or m/\%version(?:[^\w\d_]|$)/);
#print STDERR "AHEAD $_\n";
	last if $is_name_passed and $is_version_passed;
    }
    if ($is_name_ahead or $is_version_ahead) {
	$section->unshift_body('%define name '.$spec->get_section('package','')->get_tag('Name')."\n");
	$section->unshift_body('%define version '.$spec->get_section('package','')->get_tag('Version')."\n");
	print STDERR "fixed ahead!!!\n";
    }
}

sub alt_generic_fix_desc {
    my ($spec, $section)=@_;
    my @out;
    foreach (@{$section->get_bodyref()}) {
	    # % in description
	    s,%(\s|$),%% ,g;
	    push @out, &__kill_non_ascii($spec, $_);
    }
    $section->set_body(\@out);
}

sub __kill_non_ascii {
    my ($spec, $line)=@_;
    return $line if $spec->{'CONVERT::GENERIC::KEEPNONASCII'};
    local $_=$line;
    # some distro use iso or utf8 in description. ALT Linux use ascii only. :(
    #tr//e/;
    tr/ޜ\x98\x99/-.....eooeaiaoucououaaidieaayvnieouabEOOEAIAOUCOUOUAAIDIEAAYVNIEOUAB''''/;
    # exterminatus of the rest of the non-ascii characters
    s/[\x80-\xff]/./g;
    return $_;
}

sub alt_generic_check_zip {
    my ($spec)=@_;
    my $macros=$spec->macros();
    foreach my $srcitem (@{$spec->raw_sourcelist_ref}) {
	next unless $srcitem;
	$srcitem=$macros->macro_subst($srcitem);
	if ($srcitem=~/\.(?:zip|oxt|xpi)$/i) {
	    my $mainsec=$spec->get_section('package','');
	    $mainsec->unshift_body(q!# one of the sources is a zip file
BuildRequires: unzip!."\n");
	    return;
	}
    }
    foreach my $secname (qw!prep build install check!) {
	my $sec=$spec->get_section($secname);
	next unless $sec;
	if ($sec->match_body(qr'unzip\s')) {
	    my $mainsec=$spec->get_section('package','');
	    $mainsec->unshift_body(q!# required for !.$secname.q!
BuildRequires: unzip!."\n");
	}
    }
}

sub alt_generic_check_cxx {
    my ($self,$spec)=@_;
    my $mainsec=$spec->get_section('package','');
    my (@libsreq,@hdrsreq,@pkgconfreq, $flag_cxx,$flag_python);
    my $macros=$spec->macros();
    foreach my $srcitem (@{$spec->raw_sourcelist_ref}) {
	next unless $srcitem;
	$srcitem=$macros->macro_subst($srcitem);
	my $filename=$spec->{SOURCEDIR}.'/'.$srcitem;
	print STDERR "alt_generic_check_cxx: no $filename:\n" if not -e $filename and $RPM::Source::Transform::verbose>1;
	return unless -e $filename;

	my @filelist;
	my ($extractcmd,@unpackcmd);
	if ($srcitem=~/\.(?:tgz|tbz2?|txz|tar\.(?:bz2?|gz|Z|xz))$/i) {
	    $extractcmd= 'tar tf';
	    @unpackcmd= ('tar','xf');
	} elsif ($srcitem=~/\.zip$/i) {
	    $extractcmd= 'zipinfo -1';
	    @unpackcmd= ('unzip');
	}
	if ($extractcmd) {
	    open my $fn, "$extractcmd $filename |" || die "can't run $extractcmd $filename: $!";
	    while (<$fn>) {
		chomp;
		push @filelist, $_;
	    }
	    close $fn;
	}
	my @configurac=grep {m!(?:^|/)configure.(?:ac|in)$!} @filelist;
	# temporary?; let us analyze root only
	@configurac=grep {m!^(?:|\./)configure!} @configurac;
	if (grep {/\.c(?:c|pp|xx)$/ || m!(?:^|/)CMakeLists.txt$!} @filelist) {
	    $flag_cxx=1;
	}
	if (grep {m!(?:^|/)__init__.py$!} @filelist) {
	    $flag_python=1;
	}
	if (@unpackcmd and @configurac) {
	    # TODO: unpack configure.[ac|in]
	    my $curdir = getcwd;
	    my $tmpdir = File::Temp->newdir();
	    chdir $tmpdir;
	    system (@unpackcmd, $filename, @configurac)==0 or warn "unpack of $filename failed";
	    foreach my $configuracfile (@configurac) {
		open my $fn, '<', $configuracfile || die "can't open $configuracfile: $!";
		while (<$fn>) {
		    chomp;
		    if (/^\s*AC_CHECK_LIB\s*\(\s*\[?\s*([^\],\s]+)\s*[\],]/) {
			push @libsreq, $1;
		    } elsif (/^\s*AC_CHECK_HEADER\s*\(\s*\[?(\s*[^\],\s]+)\s*/) {
			push @hdrsreq, $1;
		    } elsif (/^\s*AC_PROG_CXX/) {
			$flag_cxx=1;
		    } elsif (/^\s*PKG_CHECK_MODULES\s*\(\s*\S+\s*,\s*([^\]),]+)[,)\]].*\)\s*$/) {
			my $pkgline=$1;
			$pkgline=~s/^\s*\[\s*//;
			$pkgline=~s/\s*\]\s*$//;
# note 3 arg
#PKG_CHECK_MODULES([ENCHANT], [enchant], [AC_DEFINE(HAVE_LIBENCHANT, , Define if we use enchant)], [])
# TODO: multiline (see assogiate)
#PKG_CHECK_MODULES([ASSOGIATE], [                                                
#        glib-2.0 >= 2.8.0
#        glibmm-2.4 >= 2.8.0
#        gtkmm-2.4 >= 2.6.0
#        libxml++-2.6
#        gnome-vfsmm-2.6
#])                                                                              
# TODO: variable (see xqf); note: it is inside if; so it still might be better to ignore
#        pkg_modules="gtk+-2.0 >= 2.0.0 gdk-pixbuf-xlib-2.0"
#        PKG_CHECK_MODULES(PACKAGE, [$pkg_modules])
			my @raw_pkgconf = split (/\s+/, $pkgline);
			while (@raw_pkgconf>0) {
			    my $in=shift @raw_pkgconf;
			    next unless $in and $in !~ /^\$/;
			    if ($in =~ /^[<>=]/) {
				shift @raw_pkgconf;
				next;
			    }
			    push @pkgconfreq, $in;
			}
		    }
		}
		close $fn;
	    }
	    chdir $curdir;
	    remove_tree($tmpdir);
	}
    }
    @libsreq = &__unique_list(@libsreq);
    @hdrsreq = &__unique_list(@hdrsreq);
    @pkgconfreq = &__unique_list(@pkgconfreq);
    print STDERR "BR_libs(detected):", join(' ',@libsreq),"\n" if @libsreq and $RPM::Source::Transform::verbose >1;
    print STDERR "BR_pkgconfig(detected):", join(' ',@pkgconfreq),"\n" if @pkgconfreq and $RPM::Source::Transform::verbose >1;
    my $distromap_dest=$self->get_distromap->dest();
    my @BR=map {$distromap_dest->libname2binpkgname($_)} @libsreq;
    push @BR, map {$distromap_dest->header2binpkgname($_)} @hdrsreq;
    push @BR, map {"pkgconfig($_)"} @pkgconfreq;
    @BR = sort &__unique_list(@BR);
    print STDERR "BR(detected):", join(' ',@BR),"\n" if $RPM::Source::Transform::verbose >1;
    $mainsec->unshift_body(q!# guessed from configure.ac
BuildRequires: !,join(' ',@BR)."\n") if @BR;
    $mainsec->unshift_body(q!BuildRequires: gcc-c++!."\n") if $flag_cxx;
    $mainsec->unshift_body(q!BuildRequires: python-devel!."\n") if $flag_python;
}

sub __unique_list {
    my %req=(map {$_=>1} @_);
    return keys %req;
}

sub alt_generic_check_jni {
    my ($spec)=@_;
    if (grep {/\%\{?_jnidir\}?/} map {@{$_->get_bodyref()}} $spec->get_sections()) {
	my $mainsec=$spec->get_section('package','');
	$mainsec->unshift_body(q!# required for %%_jnidir macros
BuildRequires: rpm-build-java!."\n");
    }
}

#error: Group field must be present in package: <subpackage>
sub alt_generic_merge_main_group_if_group_is_missing { 
    my ($spec,$section)=@_;
    my $group=$section->get_tag('Group');
    unless ($group) {
	my $mainsec=$spec->get_section('package','');
	my $maingroup=$mainsec->get_tag('Group');
	die "FATAL: main section does not contain a Group: tag" unless $maingroup;
	$section->set_tag('Group',$maingroup);
    }
    
}

sub alt_generic_cleanup_files {
    my $section=shift;
    my %docdir;
    my @out;
    # ALTLinux xmkmf does not create .1x man pages.
    $section->subst_if(qr'\.1x','.1',qr'^\%\{?_mandir\}?/man1/');

    foreach (@{$section->get_bodyref()}) {
	s/^\s*\%attr\s*\(-,\s*root,\s*root\s*\)\s*//;
	#if (/^\s*\%defattr\(0644,root,root,0755\)\s*$/
	if (/^\s*\%defattr\s*\((?:0644|-),\s*root,\s*root,\s*(?:0755|-)\)\s*$/
	    or /^\s*\%defattr\s*\((?:0664|-),\s*root,\s*root,\s*(?:0755|-)\)\s*$/
	    or /^\s*\%defattr\s*\((?:0644|-),\s*root,\s*root\)\s*$/
	    ) {
	    # throw away
	} elsif (/^\s*\%defattr\s*\(0755,\s*root,\s*root,\s*0755\)\s*$/) {
	    # dangerous: as in antlr may cause lost 755 rights in /usr/bin
	    push @out, "#".$_;
	} elsif (/^\s*\%attr\s*\([24](?:755|751|551),\s*root,/) {
	    # sisyphus_check: unsafe suid/sgid file permissions (readable or writable)
	    s/\%attr\s*\(2755,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(2751,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(2551,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(4755,\s*root,/\%attr\(4711,root,/;
	    s/\%attr\s*\(4751,\s*root,/\%attr\(4711,root,/;
	    s/\%attr\s*\(4551,\s*root,/\%attr\(4711,root,/;
	    push @out, $_;
	} elsif (/^\s*\%doc\s*\%\{_docdir\}\s*$/) {
	    push @out, '%doc %{_docdir}/*'."\n";
	} elsif (m!^\s*\%doc\s*(\%\{?_docdir\}?/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    push @out, $_;
	} elsif (m!^\s*(?:\%doc\s*)?(\%\{?_datadir\}?/doc/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    push @out, $_;
	} elsif (m!^\s*(\%\{?_docdir\}?/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    # note: no need to add a %doc here, rpm will add it for us.
	    push @out, $_;
	} elsif (m!^\s*((?:\%lang\(\S+\)\s*)?\%\{?_mandir\}?/\S+)\.gz\s*$!) {
	    # man pages can be auto-bzipped
	    push @out, $1.".*\n";
	} elsif (m!^\s*(\%\{?_infodir\}?/\S+)\.gz\s*$!) {
	    # info can be auto-bzipped
	    push @out, $1.".*\n";
	} elsif (m!^\s*\%{?_datadir}?/icons/\*\s*$!) {
	    #sisyphus_check: check-intersects ERROR: intersections with system packages /usr/share/icons/hicolor/48x48/apps
	    push @out, "\%_iconsdir/*/*/*/*\n";
	} elsif (m!^\s*(\%exclude)?\s*%{_libdir}/\*\.la\s*$!) {
	    # skip the la
	} else {
	    push @out, $_;
	}
    }
    $section->set_body(\@out);
    if (scalar %docdir) {
	$section->push_body(
'# hack; explicitly added docdir if not owned
'.join ('',map {'%doc %dir '.$_."\n"} keys (%docdir))
);
    }
}

sub __add_docdir {
    my ($dochashptr,$dir) =@_;
    $dir=~s/\%name-\%version/%{name}-%{version}/;
    $dir=~s/\%_docdir/%{_docdir}/;
    $dir=~s!\%{_datadir}/doc!%{_docdir}!;
    $dir=~s!\%_datadir/doc!%{_docdir}!;
    $dochashptr->{dirname($dir)}=1;
    $dochashptr->{'%{_docdir}/%{name}-%{version}'}=1;
}

sub alt_generic_fix_scripts {
    my $spec=shift;
    my $section=shift;
    my $type=$section->get_type;
    my $pkg=$section->get_canonical_package;
    my $shell=$section->get_flag('-p');
    if (defined $shell and $shell=~/ldconfig/) {
	print STDERR 'Generic2ALT: deleted %'.$type." ".$pkg." -p *ldconfig*\n";
	$section->delete();
	return;
    }
    $section->subst(qr'/sbin/ldconfig',': ');
    $section->subst(qr'\%update_maven_depmap',':');

    # tex policy
    $section->subst(qr'(?:/usr/bin/|/sbin/|/usr/sbin/|\%\{?_s?bindir\}?/?)?texhash([^;]+|$)',': ');
    $section->multi_exclude_body(
	qr'^\s*conffile="\$\(texconfig',
	qr'^\s*if\s*\[\s+"?\$1"?\s+(?:[-]eq|=)\s+"?\d+"?\s+]\s*;\s+then\s*$',
	qr'^\s*.*updmap',
	qr'^\s*fi\s*$',
	);

    # obsolete-call-in-post-gtk-update-icon-cache.pl (repocop-unittest-rpm-filetriggers-0.20-alt7)
    $section->exclude(qr'^\s*(?:/bin/)?touch\s+--no-create\s+(\%\{?_datadir\}?/icons|\%\{?_iconsdir\}?)/hicolor');
    $section->multi_exclude_body(
	qr'^\s*if\s+\[\s+-x\s+.*gtk-update-icon-cache\s+\]\s*;\s*then',
	qr'^\s*.*gtk-update-icon-cache',
	qr'^\s*fi',
	);
    $section->multi_exclude_body(
	qr'^\s*\[\s+"?\$1"?\s+-eq\s+1\s+\] &&\s+\\$',
	qr'^\s*.*gtk-update-icon-cache',
	);

    # update-desktop-database|install-info|scrollkeeper-update (obsolete in alt)
    $section->subst(qr'^\s*(?:/usr/bin/|/sbin/|/usr/sbin/|\%\{?_s?bindir\}?/?)?(update-desktop-database|install-info|scrollkeeper-update)(\s|\$)[^;]*','');

    $section->multi_exclude_body(
	qr'^\s*if\s+\[\s+-x\s+.*fc-cache\s+\]\s*;\s+then',
	qr'^\s*.*fc-cache',
	qr'^\s*fi',
	);

    #eclipse-reconciler.sh %{_libdir}/eclipse %{_tmppath} > /dev/null
    unless ($type eq 'post' and $pkg eq 'platform') { # eclipse-platform
	$section->exclude(qr'^\s*eclipse-reconciler.sh\s+\%{?_libdir}?/(?:\%{name}|eclipse)');
	print STDERR 'in %'."$type $pkg: should kill eclipse-reconciler.sh\n" if $section->match_body('eclipse-reconciler.sh');
    }

    my @body=@{$section->get_bodyref()};
    my @new_body;
    my $multiline_gtk_update_icon_cache=0;
    while (scalar @body) {
	$_=shift(@body);
	if (/gtk-update-icon-cache/) {
	    $_=&read_multiline($_, \@body) if /\\\s*$/;
	    $multiline_gtk_update_icon_cache=1;
	}
	push @new_body, $_;
    }
    if ($multiline_gtk_update_icon_cache) {
	$section->set_body(\@new_body);
    }
    $section->exclude(qr'gtk-update-icon-cache');

    if ($shell && $shell=~/gtk-update-icon-cache/) {
	$section->delete();
	return;
    }
    #---------------------------------------------------------------------

    #if [ $1 -eq 0 ]; then
    #if [ "$1" = "0" ]; then
    #fi

    $section->multi_exclude_body(
	qr'^\s*if\s*\[\s+"?\$1"?\s+(?:[-]eq|=)\s+"?\d+"?\s+]\s*;\s+then\s*$',
	qr'^\s*$',
	qr'^\s*fi\s*$',
	);
    $section->multi_exclude_body(
	qr'^\s*if\s*\[\s+"?\$1"?\s+(?:[-]eq|=)\s+"?\d+"?\s+]\s*;\s+then\s*$',
	qr'^\s*fi\s*$',
	);

    {
	# hack: filter to get rid of : as placeholders
	my @body=@{$section->get_bodyref()};
	my @new_body;
	# header
	push @new_body, shift @body;
	my $no_meaningful_lines=1;
	while (scalar @body) {
	    $_=shift(@body);
	    next if /^\s*:[^;]*$/ && $no_meaningful_lines;
	    # [ $1 -eq 0 ] || exit 0
	    next if /^\s*\[\s+"?\$1"?\s+(?:[-]eq|=)\s+"?\d+"?\s+]\s*\|\|\s*exit\s+0\s*$/ && $no_meaningful_lines;
	    # exit 0
	    next if /^\s*exit\s+0\s*$/ && $no_meaningful_lines;
	    $no_meaningful_lines=0 unless /^\s*$/ or /^\s*#/;
	    push @new_body, $_;
	}
	$section->set_body(\@new_body);
    }

    # below is just a bunch of hacks :(
    &RPM::Source::Convert::Generic2ALT::_generic_filter_script_bin_true($section);
    $section->multi_exclude_body(
qr'^\s*if\s+\[\s+-x\s+\%{_bindir}/rebuild-gcj-db\s+\]\s*$',
qr'^\s*then\s*$',
#%{_bindir}/rebuild-gcj-db
qr'/rebuild-gcj-db\s*$',
qr'^\s*fi\s*$');
    $section->exclude(qr'^\s*if\s+\[\s+-x\s+\%\{?_bindir\}?/rebuild-gcj-db\s+\];\s*then\s+\%\{?_bindir\}?/rebuild-gcj-db\s*\|\|\s*true\s*;\s*fi\s*$');
    $section->multi_exclude_body(
qr'^\s*if\s+\[\s+"\$1"\s+=\s+"0"\s+\];\s*then\s*$',
qr'^\s*:\s*$',
qr'^\s*fi\s*$');
#{
#  [ $1 -eq 0 ] || exit 0
#OR  [ $1 = 0 ] || exit 0
#  :
#} >/dev/null 2>&1 || :
    $section->multi_exclude_body(
qr'^\s*\{\s*$',
qr'^\s*\[\s+"?\$1"?\s+(?:-eq|=)\s+"?0"?\s+\]\s*\|\|\s*exit\s+0\s*$',
qr'^\s*:\s*$',
qr'^\s*\}\s*\>\s*/dev/null\s+2\>\&1\s*\|\|\s*:\s*$');
    # hack again!!!
    $section->set_body(&RPM::Source::SpecSection::__filter_empty_macro_if_endif($section->get_bodyref()));
    &RPM::Source::Convert::Generic2ALT::_generic_filter_script_bin_true($section);
    print STDERR "dead 6: %",$section->get_type()," ",$section->get_raw_package(),"\n" unless (@{$section->get_bodyref()});

    if ($section->is_empty()) {
	print STDERR 'Generic2ALT: deleted empty %'.$section->get_type()." ".$section->get_raw_package()."\n";
	$section->delete();
    } else {
	print STDERR 'Generic2ALT: keep %'.$section->get_type()." ".$section->get_raw_package()."\n";
    }
}

sub _generic_filter_script_bin_true {
    my ($section)=@_;
    my @header;
    my @out;
    my $i = 0;
    my $command_found=0;
    unless (@{$section->get_bodyref()}) {
	print STDERR "strange dead section: %",$section->get_type()," ",$section->get_raw_package(),"\n";
	return;
    }
    push @header, shift @{$section->get_bodyref()};
    foreach (@{$section->get_bodyref()}) {
	if (/^\s*:\s*$/) {
	    push @out, $_ if $i++==0 and $command_found;
	}elsif (/^\s*#/) {
	    push @out, $_;
	}elsif (/^\s*$/) {
	    push @out, $_;
	} else {
	    push @out, $_;
	    #print STDERR "other for ".$section->get_type()," ",$section->get_raw_package(),"\n",$_;
	    $command_found=1;
	    $i=0;
	}
    }
    if ($command_found) {
	$section->set_body([@header,@out]);
	#print STDERR "set cmd for ",@header;
    } else {
	$section->set_body([@header]);
	#print STDERR "Jpackage2ALT: set empty for ",@header;
    }
}




sub alt_generic_convert_alternatives {
    my $spec=shift;
    my $section=shift;
    my @out;
    my $ptr = $section->get_bodyref();
    while (@$ptr>0) {
	my $line = shift @$ptr;
	if ($line=~/\s*#/) {
	    push @out, $line;
	} elsif ($line=~s,^(\s*)(?:(?:\%{_sbindir}|/usr/sbin)/)?(?:update-)?(?:\%\{__)?alternatives\}? --(\w+),,) {
	    my $preline=$1;
	    my $updcmd=$2;
	    push @out, &_convert_single_alternative($spec,$section,$ptr,$preline,$updcmd,$line);
	} elsif ($line=~s,^(\s*)(?:(?:\%{_sbindir}|/usr/sbin)/)?(?:update-)?(?:\%\{__)?alternatives\}?\s+\\\n$,,) {
	    my $preline=$1;
	    $line = shift @$ptr;
	    if ($line=~s/^\s*--(\w+)//) {
		my $updcmd=$1;
		push @out, &_convert_single_alternative($spec,$section,$ptr,$preline,$updcmd,$line);
	    } else {
		unshift @$ptr, $line;
		push @out, $preline."# unknown update-alternatives\n";
	    }
	} else {
	    push @out, $line;
	}
    }
    $section->set_body(\@out);
}

sub _convert_single_alternative {
    my ($spec,$section,$ptr,$preline,$updcmd,$line)=@_;
    $line=&read_multiline($line, $ptr);
    # the end is abandoned :(
    if ($updcmd eq "remove") {
	my $name=&read_shell_token(\$line);
	# pre 0.4 alt alternatives
	#return $preline.'%unregister_alternatives '.&get_alternatives_file_name($spec,$section,$name)."\n";
	return $preline.":\n";
    } elsif ($updcmd eq "install") {
	my $link=&read_shell_token(\$line);
	my $name=&read_shell_token(\$line);
	my $path=&read_shell_token(\$line);
	my $priority=&read_shell_token(\$line);
	unless ($priority) {
	    $priority=10;
	    warn "alternative priority is 0 - set to $priority\n";
	}
	my @alternatives=($link,$path,$priority);
	&read_space(\$line);
	while ($line=~/--slave/) {
	    &read_shell_token(\$line);
	    my $slink=&read_shell_token(\$line);
	    my $sname=&read_shell_token(\$line);
	    my $spath=&read_shell_token(\$line);
	    push @alternatives, ($slink,$spath,$path);
	}
	&create_alternatives($spec,$section, &get_alternatives_file_name($spec,$section,$name),\@alternatives);
	return $preline.":\n";
    } else {
	warn "unknown update-alternatives --$updcmd";
	return $preline."# unknown update-alternatives --$updcmd\n";
    }
}

sub create_alternatives {
    my ($spec, $section, $alternatives_file_name,$alternativesptr) = @_;
    my @text=('install -d $RPM_BUILD_ROOT/%_altdir; cat >$RPM_BUILD_ROOT/%_altdir/'.$alternatives_file_name.'<<EOF
');
    while (@$alternativesptr>0) {
	my $provided=shift(@$alternativesptr);
	my $alternative=shift(@$alternativesptr);
	my $weight=shift(@$alternativesptr);
	push @text, $provided."\t".$alternative."\t".$weight."\n";
    }
    push @text, "EOF\n";
    $spec->get_section('install')->push_body(@text);
    my $filesec=$spec->get_section('files',$section->get_raw_package());
    if ($filesec) {
	$filesec->unshift_body('%_altdir/'.$alternatives_file_name."\n");
    } else {
	warn "WARNING: section %files for [".$section->get_raw_package()."] not found: can't write alternatives found in ".$section->get_type()."!";    }
}

sub get_alternatives_file_name {
    my ($spec,$section,$name)=@_;
#    return $name.'_'.$section->get_package_name();
    my $pkg=$section->get_canonical_package();
    return $name.($pkg ? "_$pkg" : '');
}

sub read_multiline {
    my ($line, $ptr) =@_;
    while (@$ptr>0 && $line=~s/\\[\r\n]+?$//) {
	chomp $line;
#	print "MULTILINE:$line\n";
	$line.=shift(@$ptr);
    }
    return $line;
}

sub read_shell_token {
    my ($lineptr) =@_;
    if ($$lineptr=~s/^\s*(\S+)//) {
	return $1;
    }
    return "";
}

sub read_space {
    my ($lineptr) =@_;
    $$lineptr=~s/^\s*//;
}

1;
