#!/usr/bin/perl -w

package RPM::Source::Convert::Mass;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use File::Path qw(make_path remove_tree);
use RPM::Header;
use RPM qw/evrcmp/;
use DistroMap 0.05;
use RPM::Source::Transform 0.002;

use String::ShellQuote;

require Exporter;

our $VERSION=0.05;

our @ISA = qw(Exporter);
our @EXPORT = qw/$DESTPATH $ORIGINPATH @ORIGINSKIP @DESTSKIP %ORIGIN2DESTMAP/;
our ($DESTPATH, $ORIGINPATH);
my (@DESTPATH, @ORIGINPATH);

our @ORIGINSKIP=();
our @DESTSKIP=();
our %ORIGIN2DESTMAP=();
my %DEST2ORIGINMAP=();
my %ORIGINSKIP;
my %DESTSKIP;

our $OUTDIR="./OUT";

our $origintoolsdir=dirname($0);

our $verbose=0;
#our ($changelog, @hooks);
our ($skip_conversion_errors, $help, $parent);
our $strategy = 'origin2dest';
our $originimport="srpmconvert";
our $originmode=1;
our $suppress_rename=0;
my ($opt_transaction, $opt_taskfile, $opt_namefile);
our $arg_timestampdir;
our $arg_timestamp_expire=7; # in days


our %STRATEGYMAP=(
    'all' => \&origin2dest,
    'missing' => \&originskipdest,
    'orphaned' => \&destskiporigin,
    'destup' => \&destup,
    'newer' => \&newer,
    'older' => \&older,
    # old names
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destskiporigin' => \&destskiporigin,
    );

our %STRATEGYDESCR=(
    'all' => 'any package',
    'missing' => 'packages in origin, missing in dest', 
    'orphaned' => 'packages in dest, missing in origin',
    'destup' => 'packages in dest, older than origin (i.e. need update)',
    'newer' => 'packages in origin, new and newer than dest',
    'older' => 'packages in origin, older than dest',
    );

sub new {
    my $class = shift;
    my $self = {
	'-cliargs' => [],
	'-packages' => [],
	@_
    };
    bless $self, $class;
    my $result = GetOptions (
	&RPM::Source::Transform::Getopt::PassThrough::passthrough(@RPM::Source::Transform::PASSTHROUGH_OPTIONS),
#	"changelog=s"	=> \$changelog,
#	"hook=s"	=> \@hooks,
#	"parent=s"	=> \$parent,
	"dest=s"	=> \$DESTPATH,
	"help"		=> \$help,
	"input-dest"	=> sub {$originmode=0},
	"input-origin"	=> sub {$originmode=1},
	"origin=s"	=> \$ORIGINPATH,
	"out|outdir=s"	=> \$OUTDIR,
	"skip-errors"	=> \$skip_conversion_errors,
	"strategy=s"	=> \$strategy,
	"timestampdir=s"	=> \$arg_timestampdir,
	"timestamp-expire=f"	=> \$arg_timestamp_expire,
	"transaction"	=> \$opt_transaction,
	"taskfile|tasklist=s"	=> \$opt_taskfile,
	"namefile|namelist=s"	=> \$opt_namefile,
	"v|verbose+"	=> \$verbose,
	'q|quiet'   => sub { $verbose = 0 },
	@DistroMap::LONGOPT,
	@{$self->{'-cliargs'}},
	);

    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }

    $self->{VERBOSE}=$verbose;
    $self->process_options_hook();

    @DESTPATH=&path_to_array($DESTPATH);
    @ORIGINPATH=&path_to_array($ORIGINPATH);
    die "invalid origin path or destination path\n" if ((! scalar @DESTPATH) || (! scalar @ORIGINPATH));

    # DEPRECATED: REMOVE ...
    my @distromap_args;
    foreach my $arg ('-originrepo', '-destrepo','-originbranch', '-destbranch') {
	push @distromap_args, $arg => $self->{$arg} if $self->{$arg};
    }
    $self->{'__distromap'}=DistroMap->new(@distromap_args);

    %ORIGINSKIP=map {$_ => 1} @ORIGINSKIP;
    %DESTSKIP=map {$_ => 1} @DESTSKIP;

    # TODO: handle both?
    %DEST2ORIGINMAP=reverse (%ORIGIN2DESTMAP);

    push @{$self->{'-packages'}}, $opt_transaction ? [@ARGV] : @ARGV;

    my @all_names;
    if ($opt_namefile) {
	open (my $fn, '<', $opt_namefile) or die "can't open file $opt_namefile\n";
	while (my $line=<$fn>) {
	    my @list = &__line2array($line);
	    push @all_names, $list[0];
	}
	close($fn);
    }
    if ($opt_taskfile) {
	open (my $fn, '<', $opt_taskfile) or die "can't open file $opt_taskfile\n";
	while (my $line=<$fn>) {
	    my @list = map {&__expand_glob($_,\@all_names)} &__line2array($line);
	    next if scalar @list == 0;
	    if (scalar @list == 1) {
		push @{$self->{'-packages'}}, @list;
	    } else {
		# transaction
		push @{$self->{'-packages'}}, [@list];
	    }
	}
	close($fn);
    }

    make_path($arg_timestampdir) if defined $arg_timestampdir;
    return $self;
}

# glob expansion on $opt_namefile
sub __expand_glob {
    my ($pattern, $arrayptr)=@_;
    return $pattern unless $pattern =~/[?*]/;
    my $regexp=&glob2pat($pattern);
    my @names = grep {/$regexp/} @$arrayptr;
    print STDERR "WARNING: pattern $pattern does not match.\n" unless @names;
    return @names;
}

sub __line2array {
    my $line=shift;
    chomp $line;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    return if $line=~/^\s*#/;
    return if $line eq '';
    return split (/\s+/, $line);
}

sub process_options_hook {
    my ($self)=@_;
}

sub usage {
    print "usage: $0 [ options ] [ list of origin rpms/names ]

Example:
upgrade all dest java packages using jpackage
originmass --dest /var/ftp/pub/Sisyphus/files/SRPMS --origin /var/ftp/pub/jpackage/1.7/generic/free/SRPMS

General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet

Mass processing options:
	--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
	--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
	--out	/path/to/output_dir (default is $OUTDIR)
	--skip-errors	ignore individual failures
	--strategy 'strategy'	of mass operations (default is $strategy).
	  Available strategies: 
";
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	print "\t\t$sname\t- $description\n" if defined $description;
    }

    print "\n\n";
    &DistroMap::longopt_usage();
    print "
The following editing options are passed through to the underlying command.
Pass-through options:
";
    &RPM::Source::Transform::usage_passthrough_options();
}

sub new_mapping {
    my ($self, $name)=@_;
    return RPM::Source::Convert::Mass::Mapping->new('__distromap'=>$self->{__distromap},'-name'=>$name);
}

sub run {
    my ($self)=@_;
    foreach my $nameobj (@{$self->{'-packages'}}) {
	if (ref $nameobj) {
	    $self->run_transaction($nameobj);
	} else {
	    $self->run_name($nameobj);
	}
    }
}

sub run_name {
    my ($self, $name)=@_;
    my $rpmmap=$self->new_mapping($name);
    return if $self->__skip_due_to_timestamp($rpmmap);
    return unless $self->is_strategy_allow($rpmmap);
    $self->__mk_timestamp($rpmmap);
    $self->run_origin_import($rpmmap);
}

sub run_transaction {
    my ($self, $transaction_ptr)=@_;
    my $transaction_name=$transaction_ptr->[0];
    my $is_in_transaction=0;
    my $transaction_count=0;
    foreach my $name (@$transaction_ptr) {
	my $rpmmap=$self->new_mapping($name);
	next if $self->__skip_due_to_timestamp($rpmmap);
	next unless $self->is_strategy_allow($rpmmap);
	if (not $is_in_transaction) {
	    $self->init_transaction($transaction_name);
	    $is_in_transaction=1;
	}
	$self->__mk_timestamp($rpmmap);
	$self->run_origin_import($rpmmap, $transaction_name, ++$transaction_count);
    }
    $self->done_transaction($transaction_name,$transaction_count) if $is_in_transaction;
}

sub init_transaction {
    my ($self,$transaction_name)=@_;
}

sub done_transaction {
    my ($self,$transaction_name)=@_;
}

sub is_strategy_allow {
    my ($self,$rpmmap)=@_;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    die "strategy $strategy not found" unless $strategy_allow;
    my $originname=$rpmmap->originname;
    my $destname=$rpmmap->destname;
    return 0 if $ORIGINSKIP{$originname} or $DESTSKIP{$destname};
    return 1 if &{$strategy_allow}($self,$originname, $destname,$rpmmap->origin_rh, $rpmmap->dest_rh);
    return 0;
}

sub __skip_due_to_timestamp {
    my ($self,$rpmmap)=@_;
    return 0 unless defined $arg_timestampdir;
    my $timestamp_file=&__srpmfile2timestamp($rpmmap->origin_rh->source_name);
    return 1 if -M $timestamp_file and ($arg_timestamp_expire==0 or -M $timestamp_file < $arg_timestamp_expire);
    return 0;
}

sub __mk_timestamp {
    my ($self,$rpmmap)=@_;
    return unless defined $arg_timestampdir;
    &__run('touch',&__srpmfile2timestamp($rpmmap->origin_rh->source_name));
}

sub run_origin_import {
    my ($self,$rpmmap,$transaction_name,$transaction_count)=@_;
    make_path ($OUTDIR) unless -d $OUTDIR;
    my ($originname, $destname, $originrpm, $destrpm)=($rpmmap->originname,$rpmmap->destname,$rpmmap->originrpm,$rpmmap->destrpm);
    my @args;
    if ($verbose) {
	for (my $i=0; $i<$verbose-1; $i++) { # less verbose by 1; assumed verbose=1 is for mass only
	    push @args, '--verbose';
	}
    }

    push @args, '--rename', $destname if $destname ne $originname and not $suppress_rename;

    push @args, '--outdir', defined($transaction_name) ? "$OUTDIR/$transaction_name.transaction/".('0' x (5-length($transaction_count))).$transaction_count : $OUTDIR;

#    push @args, '--changelog', $changelog if ($changelog);

    if ($parent) {
	push @args, '--parent', $parent;
    } else {
	push @args, '--parent-rpm', $destrpm if ($destrpm);
    }
    if ($originrpm) {
	push @args, '--rpm', $originrpm;
    } else {
	print STDERR "FATAL: foreign RPM $originname not found.\n";
	return;
    }
    push @args, &RPM::Source::Transform::Getopt::PassThrough::cat_options();

    my $hook;
    $hook="$origintoolsdir/hooks/$originname.pl";
    push @args, '--hook', $hook, '--no-defaulthook' if (-e $hook);
    #foreach $hook (@hooks) {
    #	push @args, '--hook', $hook;
    #}
    $self->origin_import_hook(\@args, $rpmmap);

    print join(' ',map {shell_quote_best_effort $_} $originimport,@args),"\n";
    if (system($originimport,@args)!=0) {
	print STDERR "ERROR: $originimport failed\n";
	die "exiting.\n" unless $skip_conversion_errors;
    }
    print "---*---\n";
}

sub origin_import_hook {}

sub origin_cmp_dest {
    my ($self, $origin_rh, $dest_rh)=@_;
    return 0 unless $origin_rh;
    return 1 unless $dest_rh;
    return $origin_rh->cmpver($dest_rh);
}

#------------------------------------------

#------------------------------------------
#    private strategies
#------------------------------------------

sub origin2dest {
    my ($self,$originname, $destname,$origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin_rh;
    return 0 unless $origin_rh;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest_rh;
    return 1;
}

sub destup {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin_rh;
    return 0 unless $origin_rh;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest_rh;
    return 1 if $dest_rh and $self->origin_cmp_dest($origin_rh,$dest_rh)>0;
    return 0;
}

sub newer {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin_rh;
    return 0 unless $origin_rh;
    print STDERR "note: origin RPM $originname not found.\n" unless $origin_rh;
    return 1 if not $dest_rh or $self->origin_cmp_dest($origin_rh,$dest_rh)>0;
    return 0;
}

sub originskipdest {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin_rh;
    return 0 unless $origin_rh;
    print STDERR "found dest RPM $destname for $originname.\n" if $dest_rh;
    return 0 if $dest_rh;
    return 1;
}

sub older {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin_rh;
    return 0 unless $origin_rh;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest_rh;
    return 1 if $dest_rh and $self->origin_cmp_dest($origin_rh,$dest_rh)<0;
    return 0;
}

sub destskiporigin {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "FATAL: dest RPM $destname not found.\n" unless $dest_rh;
    return 0 unless $dest_rh;
    print STDERR "found origin RPM $originname for $destname.\n" if $origin_rh;
    return 0 if $origin_rh;
    return 1;
}


#------------------------------------------

#------------------------------------------
#    private utilities
#------------------------------------------

sub path_to_array {
    my $paths=shift;
    my @path;
    foreach (split(':',$paths)) {
	if ($_ eq '') {
	    $_ = '.';
	} elsif (/^~/) {
	    s/^~/$ENV{HOME}/;
	}
	push @path, $_ if -d $_;
    }
    return @path;
}

sub glob2pat {
    my $globstr = shift;
    my %patmap = (
	'*' => '.*',
	'?' => '.',
	'[' => '[',
	']' => ']',
	'{' => '(?:',
	'}' => ')',
	',' => '|',
	);
    $globstr =~ s{(.)} { $patmap{$1} || "\Q$1" }ge;
    return '^' . $globstr . '$';
}


sub __srpmfile2timestamp {
    my ($rpm)=@_;
    return $arg_timestampdir.'/'.basename($rpm).'.stamp';
}

sub __run {
    print STDERR "running: ", join(' ',@_),"\n" if $RPM::Source::Convert::Mass::verbose;
    system(@_)==0 or Carp::confess "command failed: ".join(' ',@_);
}

package RPM::Source::Convert::Mass::Mapping;

our $find_origin_sub=\&RPM::Source::Convert::Mass::Mapping::find_first_rpm;
our $find_dest_sub=\&RPM::Source::Convert::Mass::Mapping::find_highest_rpm;

sub new {
    my $class = shift;
    my $self = {
	#'__distromap' => 
	@_
    };
    Carp::confess "Mass::mapping:new:undefined name" unless defined $self->{-name};
    bless $self, $class;
    # TODO:
    #if ($name=~/.src.rpm$/) {
    #$name=& MassCommon:: __get_rpm_name($rpm);
    #}
    $self->__get_dest_origin_name($self->{-name});
    return $self;
}

sub name {
    my ($self)=@_;
    return $self->{-name};
}

sub destname {
    my ($self)=@_;
    return $self->{-destname};
}

sub originname {
    my ($self)=@_;
    return $self->{-originname};
}

sub origin_rh {
    my ($self)=@_;
    return $self->{-origin_rh} if defined $self->{-origin_rh};
    $self->{-origin_rh}=$self->__find_origin_rpm($self->{-originname});
    return $self->{-origin_rh};
}

sub dest_rh {
    my ($self)=@_;
    return $self->{-dest_rh} if defined $self->{-dest_rh};
    $self->{-dest_rh}=$self->__find_dest_rpm($self->{-destname});
    return $self->{-dest_rh};
}

sub originrpm {
    my ($self)=@_;
    my $origin_rh=$self->origin_rh;
    my $originrpm;
    $originrpm=$origin_rh->source_name if $origin_rh;
    return $originrpm;
}

sub destrpm {
    my ($self)=@_;
    my $dest_rh=$self->dest_rh;
    my $destrpm;
    $destrpm=$dest_rh->source_name if $dest_rh;
    return $destrpm;
}


sub __origin2dest { 
    my ($self,$name)=@_;
    my $destname;
    if ($RPM::Source::Convert::Mass::ORIGIN2DESTMAP{$name}) {
	return $RPM::Source::Convert::Mass::ORIGIN2DESTMAP{$name};
    } else {
	my $destname=$self->{'__distromap'}->unimap_source_origin2dest($name);
	return $destname if $destname;
    }
    return $name;
}

sub __dest2origin { 
    my ($self,$name)=@_;
    my $destname;
    if ($DEST2ORIGINMAP{$name}) {
	return $DEST2ORIGINMAP{$name};
    } else {
	my $oname=$self->{'__distromap'}->unimap_source_dest2origin($name);
	return $oname if $oname;
    }
    return $name;
}

sub __get_dest_origin_name {
    my ($self,$name)=@_;
    my $destname=$name;
    my $originname=$name;
    if ($originmode) {
	my $newdest=$self->__origin2dest($originname);
	$destname=$newdest if $newdest;
    } else {
	my $neworigin=$self->__dest2origin($destname);
	$originname=$neworigin if $neworigin;
    }
    $self->{-destname} = $destname;
    $self->{-originname} = $originname;
}

sub __find_origin_rpm {
    my ($self,$name)=@_;
    return &$find_origin_sub($name,@ORIGINPATH);
}

sub __find_dest_rpm {
    my ($self,$name)=@_;
    return &$find_dest_sub($name,@DESTPATH);
}

sub find_highest_rpm {
    my $name=shift;
    my $cur_href;
    for my $path (@_) {
      PATH:
	for my $rpmfile (reverse(glob "$path/$name-*.src.rpm")) {
#	print STDERR "trying $rpmfile for $name\n";
	    my $href = new RPM::Header $rpmfile;
	    if ($name eq $href->{NAME}) {
		if (not $cur_href or $cur_href->cmpver($href)<0) {
		    $cur_href=$href;
		}
		# assume 1 rpm per name; false for jpp TODO:option!!!
		last PATH;
	    }
	}
    }
    return $cur_href;
}

sub find_first_rpm {
    my $name=shift;
    for my $path (@_) {
	for my $rpmfile (reverse(glob "$path/$name-*.src.rpm")) {
#	print STDERR "trying $rpm for $name\n";
	    my $href = new RPM::Header $rpmfile;
	    if ($name eq $href->{NAME}) {
		# assume 1 rpm per name; false for jpp TODO:option!!!
		return $href;
	    }
	}
    }
    return;
}

1;
